#include <SDL.h>

typedef struct VideoInfo {
	Sint64 numframes;
	Sint64 * frameoffsets;
	Sint64 filesize;
	double fps;
	size_t framesize;
	int w, h;
} VideoInfo;

typedef struct PlayerInfo {
	struct {
		double start;
		double current;
		double max;
	} pos;
	Sint64 currentframe;
	Sint64 bufferframe;
	Uint8 * buffer;
	int paused;
} PlayerInfo;

static SDL_Window * window;
static SDL_Renderer * renderer;
static SDL_Texture * texture_yuv;
static SDL_Texture * texture_overlay_rgb;
static SDL_Texture * texture_overlay_rgba;
static SDL_Texture * texture_surface_rgb;
static SDL_Texture * texture_surface_rgba;
static SDL_Texture * texture_for_triangle;
static SDL_Surface * surface_rgb;
static SDL_Surface * surface_rgba;
static SDL_RWops * videofile;
static PlayerInfo player;

/* Parses the yuv4mpeg2 video file */
static int parse_yuv4mpeg(SDL_RWops * file, VideoInfo * info)
{
	Uint8 tmpbuf[1024] = {0};
	size_t i, j, bytes, offsetsize;
	double denom;
	Sint64 offset;

	SDL_RWseek(file, 0, RW_SEEK_SET);

	bytes = SDL_RWread(file, tmpbuf, 1, 1023);
	tmpbuf[bytes >= 0 ? bytes : 0] = 0;
	if (bytes < 10 || SDL_memcmp(tmpbuf, "YUV4MPEG2", 9) != 0) {
		SDL_Log("Unkown file format");
		return 1;
	}

	i = 9;
	while (tmpbuf[i]) {
		if (tmpbuf[i] == '\n') {
			break;
		} else if (tmpbuf[i] != ' ') {
			SDL_Log("Unexpected character at %d", (int)i);
			return 1;
		}

		i++;
		if (i >= bytes) {
			SDL_Log("Could not read header");
			return 1;
		}

		switch (tmpbuf[i]) {
		case 'C':
			if (SDL_strncasecmp(tmpbuf + i + 1, "420jpeg", 7) != 0 && SDL_strncasecmp(tmpbuf + i + 1, "420mpeg2", 8) != 0) {
				SDL_Log("Unsupported format");
				return 1;
			}
			break;
		case 'W':
			info->w = SDL_atoi(tmpbuf + i + 1);
			break;
		case 'H':
			info->h = SDL_atoi(tmpbuf + i + 1);
			break;
		case 'F':
			j = i + 1;
			while (!(tmpbuf[j] == 0 || tmpbuf[j] == ' ' || tmpbuf[j] == '\n' || tmpbuf[j] == ':')) {
				j++;
			}
			if (tmpbuf[j] != ':') {
				SDL_Log("Malformed framerate ratio");
				return 1;
			}
			denom = (double)SDL_atoi(tmpbuf + j + 1);
			if (denom == 0.0) {
				info->fps = 30;
			} else {
				info->fps = (double)SDL_atoi(tmpbuf + i + 1) / denom;
			}
			break;
		}
		while (!(tmpbuf[i] == 0 || tmpbuf[i] == ' ' || tmpbuf[i] == '\n')) i++;
	}

	if (tmpbuf[i] != '\n') {
		SDL_Log("Unexpected end of header");
		return 1;
	}
	i++;

	if (info->w < 1 || info->h < 1) {
		SDL_Log("Invalid width or height");
		return 1;
	}
	info->framesize = info->w * info->h * 3 / 2;

	offsetsize = 1024 * sizeof(Sint64);
	info->frameoffsets = SDL_malloc(offsetsize);
	if (info->frameoffsets == NULL) {
		SDL_Log("No memory");
		return 1;
	}
	info->numframes = 0;
	offset = SDL_RWseek(file, i, RW_SEEK_SET);

	while (tmpbuf[i]) {
		i = 0;
		if (offset < 0) {
			SDL_Log("Could not get offset");
			return 1;
		}
		bytes = SDL_RWread(file, tmpbuf, 1, 128);
		if (bytes < 6) {
			break;
		}
		tmpbuf[bytes] = 0;
		if (SDL_memcmp(tmpbuf, "FRAME", 5) != 0) {
			SDL_Log("Frame missing at %d", (int)offset);
			return 1;
		}

		if (info->numframes >= (Sint64)(offsetsize / sizeof(Sint64))) {
			Sint64 * newoffsets = SDL_realloc(info->frameoffsets, offsetsize * 2);
			if (newoffsets == NULL) {
				SDL_free(info->frameoffsets);
				info->frameoffsets = NULL;
				SDL_Log("No memory");
				return 1;
			}
			offsetsize *= 2;
		}

		while (!(tmpbuf[i] == 0 || tmpbuf[i] == '\n')) i++;
		if (tmpbuf[i] != '\n') {
			SDL_Log("Unexpected character at %d", (int)i);
			return 1;
		}

		info->frameoffsets[info->numframes] = offset + i + 1;
		offset = SDL_RWseek(file, offset + i + 1 + info->framesize, RW_SEEK_SET);
		if (offset < 0) {
			SDL_Log("Could not seek to next frame");
			return 1;
		}
		info->numframes++;
	}

	if (info->numframes < 1) {
		SDL_Log("This video has no frames");
		return 1;
	}

	return 0;
}

static void SetRect(SDL_Rect * r, int x, int y, int w, int h)
{
	r->x = x;
	r->y = y;
	r->w = w;
	r->h = h;
}

static double GetNow()
{
	return (double)SDL_GetPerformanceCounter() / (double)SDL_GetPerformanceFrequency();
}

/* Draws to the overlay target texture with a renderer. */
static int UpdateOverlayTexture(SDL_Texture * texture, int paused, double now, double progress)
{
	SDL_Rect dst;

	SDL_SetRenderTarget(renderer, texture);

	SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
	SDL_RenderClear(renderer);
	SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);

	SetRect(&dst, 10, 130, 180, 10);
	SDL_RenderDrawRect(renderer, &dst);

	SetRect(&dst, 12, 132, (int)(176 * progress), 6);
	SDL_RenderFillRect(renderer, &dst);

	if (paused) {
		SetRect(&dst, 170, 15, 4, 20);
		SDL_RenderFillRect(renderer, &dst);
		SetRect(&dst, 180, 15, 4, 20);
		SDL_RenderFillRect(renderer, &dst);
	} else {
		SetRect(&dst, 162, 15, 20, 16);
		SDL_RenderCopyEx(renderer, texture_for_triangle, NULL, &dst, 45, NULL, SDL_FLIP_NONE);

		SetRect(&dst, 162 - 8, 15 - 7, 20, 35);
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
		SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE);
		SDL_RenderFillRect(renderer, &dst);
		SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
		SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND);
	}

	SDL_SetRenderTarget(renderer, NULL);

	return 0;
}

/* Draws to the overlay surface and then updates the texture with the surface data. */
static int UpdateOverlaySurface(SDL_Surface * surface, SDL_Texture * texture, int paused, double now, double progress)
{
	SDL_Rect dst;
	Uint32 white = SDL_MapRGBA(surface->format, 255, 255, 255, 255);

	SDL_FillRect(surface, NULL, SDL_MapRGBA(surface->format, 0, 0, 0, 0));

	SetRect(&dst, 10, 130, 180, 1);
	SDL_FillRect(surface, &dst, white);
	SetRect(&dst, 10, 139, 180, 1);
	SDL_FillRect(surface, &dst, white);
	SetRect(&dst, 10, 131, 1, 8);
	SDL_FillRect(surface, &dst, white);
	SetRect(&dst, 189, 131, 1, 8);
	SDL_FillRect(surface, &dst, white);

	SetRect(&dst, 12, 132, (int)(176 * progress), 6);
	SDL_FillRect(surface, &dst, white);

	if (paused) {
		SetRect(&dst, 170, 15, 4, 20);
		SDL_FillRect(surface, &dst, white);
		SetRect(&dst, 180, 15, 4, 20);
		SDL_FillRect(surface, &dst, white);
	} else {
		int y;
		for (y = 15; y < 35; y++) {
			SetRect(&dst, 170, y, y < 15 + 10 ? (y - 14) : (34 - y) , 1);
			SDL_FillRect(surface, &dst, white);
		}
	}

	SDL_UpdateTexture(texture, NULL, surface->pixels, surface->pitch);

	return 0;
}

static void UpdateWindowTitle(const Uint8 * name, Uint32 blend, int alpha)
{
	Uint8 title[256];
	SDL_snprintf(title, 256, "%s/b:%s/a:%d", name,
		blend == SDL_BLENDMODE_NONE ? "NONE" :
		blend == SDL_BLENDMODE_BLEND ? "BLEND" :
		blend == SDL_BLENDMODE_ADD ? "ADD" :
		blend == SDL_BLENDMODE_MOD ? "MOD" : "UNKNOWN",
		alpha);
	SDL_SetWindowTitle(window, title);
}

/* Main function */
static int LoadAndPlay(int argc, char**argv)
{
	int i, done = 0, yuv_available = 0, window_width, window_height;
	VideoInfo videoinfo = {0};
	SDL_RendererInfo info;
	SDL_Rect topleft, topright, bottomleft, bottomright;
	SDL_BlendMode texture_blendmode = SDL_BLENDMODE_BLEND;
	int texture_alphamod = 255;

	int surfacebpp;
	Uint32 Rmask, Gmask, Bmask, Amask, texformat;

	SetRect(&topleft, 0, 0, 200, 150);
	SetRect(&topright, 205, 0, 200, 150);
	SetRect(&bottomleft, 0, 155, 200, 150);
	SetRect(&bottomright, 205, 155, 200, 150);

	if (argc <= 1) {
		SDL_Log("Requires a filename as the first argument");
		return 1;
	}

	videofile = SDL_RWFromFile(argv[1], "rb");
	if (videofile == NULL) {
		SDL_Log("Could not open file");
		return 2;
	}

	videoinfo.filesize = SDL_RWseek(videofile, 0, RW_SEEK_END);
	if (videoinfo.filesize < 0) {
		SDL_Log("Could not determine filesize");
		return 3;
	}

	if (parse_yuv4mpeg(videofile, &videoinfo)) {
		return 4;
	}
	SDL_Log("Video: %dx%d@%.1f with %d frames and a duration of %.3f seconds.", videoinfo.w, videoinfo.h, videoinfo.fps, (int)videoinfo.numframes, videoinfo.numframes / videoinfo.fps);

	player.buffer = SDL_malloc(videoinfo.filesize);
	if (player.buffer == NULL) {
		return 6;
	}

	window = SDL_CreateWindow("YUV with overlay", 100, 100, 405, 305, 0);
	if (window == NULL) {
		return 7;
	}

	renderer = SDL_CreateRenderer(window, -1, 0);
	if (renderer == NULL) {
		return 8;
	}

	SDL_GetRendererInfo(renderer, &info);

	for (i = 0; i < (int)info.num_texture_formats; i++) {
		if (info.texture_formats[i] == SDL_PIXELFORMAT_IYUV) {
			yuv_available = 1;
		}
	}

	if (!(info.flags & SDL_RENDERER_TARGETTEXTURE)) {
		SDL_Log("Renderer doesn't support target textures");
		return 9;
	}
	if (!yuv_available) {
		SDL_Log("Renderer doesn't support YUV");
		return 10;
	}

	/* Surface creation */
	SDL_PixelFormatEnumToMasks(SDL_PIXELFORMAT_RGB888, &surfacebpp, &Rmask, &Gmask, &Bmask, &Amask);
	surface_rgb = SDL_CreateRGBSurface(0, 200, 150, surfacebpp, Rmask, Gmask, Bmask, Amask);
	if (surface_rgb == NULL) {
		SDL_Log("Could not create RGB surface");
		return 110;
	}
	SDL_PixelFormatEnumToMasks(SDL_PIXELFORMAT_ARGB8888, &surfacebpp, &Rmask, &Gmask, &Bmask, &Amask);
	surface_rgba = SDL_CreateRGBSurface(0, 200, 150, surfacebpp, Rmask, Gmask, Bmask, Amask);
	if (surface_rgb == NULL) {
		SDL_Log("Could not create RGBA surface");
		return 120;
	}

	/* Texture creation */
	texture_yuv = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_IYUV, SDL_TEXTUREACCESS_STREAMING, videoinfo.w, videoinfo.h);
	if (texture_yuv == NULL) {
		SDL_Log("Could not create YUV texture");
		return 130;
	}
	texture_overlay_rgb = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGB888, SDL_TEXTUREACCESS_TARGET, 200, 150);
	if (texture_overlay_rgb == NULL) {
		SDL_Log("Could not create RGB texture");
		return 140;
	}
	texture_overlay_rgba = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_TARGET, 200, 150);
	if (texture_overlay_rgba == NULL) {
		SDL_Log("Could not create RGBA texture");
		return 150;
	}
	texture_surface_rgb = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGB888, SDL_TEXTUREACCESS_STATIC, 200, 150);
	if (texture_surface_rgb == NULL) {
		SDL_Log("Could not create surface RGB texture");
		return 160;
	}
	texture_surface_rgba = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STATIC, 200, 150);
	if (texture_surface_rgba == NULL) {
		SDL_Log("Could not create surface RGBA texture");
		return 170;
	}
	texture_for_triangle = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STATIC, 1, 1);
	if (texture_for_triangle == NULL) {
		SDL_Log("Could not create triangle texture");
		return 180;
	} else {
		Uint32 white = 0xffffffff;
		SDL_UpdateTexture(texture_for_triangle, NULL, &white, 4);
	}

	/* The surface format has to match the texture format or it will upload garbage. */
	SDL_QueryTexture(texture_surface_rgb, &texformat, NULL, NULL, NULL);
	if (texformat != surface_rgb->format->format) {
		SDL_Log("Texture format (%s) didn't match surface format (%s)", SDL_GetPixelFormatName(texformat), SDL_GetPixelFormatName(surface_rgb->format->format));
		return 190;
	}

	SDL_GetWindowSize(window, &window_width, &window_height);

	SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
	SDL_RenderClear(renderer);
	SDL_RenderPresent(renderer);

	player.pos.start = GetNow();
	player.pos.max = videoinfo.numframes / videoinfo.fps;
	player.bufferframe = -1;

	UpdateWindowTitle(info.name, texture_blendmode, texture_alphamod);

	while (!done) {
		SDL_Event e;
		SDL_Rect dst;
		double now = GetNow();

		while (SDL_PollEvent(&e)) {
			if (e.type == SDL_QUIT) {
				done = 1;
			} else if (e.type == SDL_KEYUP) {
				Uint32 sym = e.key.keysym.sym;
				if (sym == SDLK_ESCAPE) {
					done = 1;
				} else if (sym == SDLK_SPACE) {
					player.paused = !player.paused;
					if (!player.paused) {
						player.pos.start += now - (player.pos.start + player.pos.current);
					}
				} else if (sym == SDLK_F1) {
					texture_blendmode = SDL_BLENDMODE_NONE;
				} else if (sym == SDLK_F2) {
					texture_blendmode = SDL_BLENDMODE_BLEND;
				} else if (sym == SDLK_F3) {
					texture_blendmode = SDL_BLENDMODE_ADD;
				} else if (sym == SDLK_F4) {
					texture_blendmode = SDL_BLENDMODE_MOD;
				} else if (sym == SDLK_1) {
					texture_alphamod = 0;
				} else if (sym == SDLK_2) {
					texture_alphamod = 0x33;
				} else if (sym == SDLK_3) {
					texture_alphamod = 0x77;
				} else if (sym == SDLK_4) {
					texture_alphamod = 0xaa;
				} else if (sym == SDLK_5) {
					texture_alphamod = 0xff;
				}
				UpdateWindowTitle(info.name, texture_blendmode, texture_alphamod);
			} else if (e.type == SDL_WINDOWEVENT) {
				if (e.window.event == SDL_WINDOWEVENT_RESIZED || e.window.event == SDL_WINDOWEVENT_SIZE_CHANGED) {
					window_width = e.window.data1;
					window_height = e.window.data2;
				}
			}
		}


		if (!player.paused) {
			player.pos.current = now - player.pos.start;
			if (player.pos.current > player.pos.max) {
				player.pos.start = now;
				player.pos.current = 0;
			}
		}

		/* Upload new video frame if necessary. */
		if (SDL_floor(player.pos.current * videoinfo.fps) != player.currentframe) {
			Sint64 newframe = (Sint64)SDL_floor(player.pos.current * videoinfo.fps);
			if (player.bufferframe == newframe) {
				SDL_UpdateTexture(texture_yuv, NULL, player.buffer, videoinfo.w);
			} else {
				/* Frame not in buffer. Read directly from file. */
				void * pixels;
				int pitch;
				SDL_RWseek(videofile, videoinfo.frameoffsets[newframe], RW_SEEK_SET);
				SDL_LockTexture(texture_yuv, NULL, &pixels, &pitch);
				SDL_RWread(videofile, pixels, 1, videoinfo.framesize);
				SDL_UnlockTexture(texture_yuv);
			}
			player.currentframe = newframe;
		}

		SDL_SetTextureBlendMode(texture_overlay_rgb, texture_blendmode);
		SDL_SetTextureBlendMode(texture_overlay_rgba, texture_blendmode);
		SDL_SetTextureBlendMode(texture_surface_rgb, texture_blendmode);
		SDL_SetTextureBlendMode(texture_surface_rgba, texture_blendmode);

		SDL_SetTextureAlphaMod(texture_overlay_rgb, texture_alphamod);
		SDL_SetTextureAlphaMod(texture_overlay_rgba, texture_alphamod);
		SDL_SetTextureAlphaMod(texture_surface_rgb, texture_alphamod);
		SDL_SetTextureAlphaMod(texture_surface_rgba, texture_alphamod);

		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
		SDL_RenderClear(renderer);

		/* Updating the surfaces and textures. None of the following calls will drawn to the screen. */
		UpdateOverlayTexture(texture_overlay_rgb, player.paused, now, player.pos.current / player.pos.max);
		UpdateOverlayTexture(texture_overlay_rgba, player.paused, now, player.pos.current / player.pos.max);
		UpdateOverlaySurface(surface_rgb, texture_surface_rgb, player.paused, now, player.pos.current / player.pos.max);
		UpdateOverlaySurface(surface_rgba, texture_surface_rgba, player.paused, now, player.pos.current / player.pos.max);

		/* Copy the video to the screen. */
		dst = topleft;
		SDL_RenderCopy(renderer, texture_yuv, NULL, &dst);
		dst = topright;
		SDL_RenderCopy(renderer, texture_yuv, NULL, &dst);
		dst = bottomleft;
		SDL_RenderCopy(renderer, texture_yuv, NULL, &dst);
		dst = bottomright;
		SDL_RenderCopy(renderer, texture_yuv, NULL, &dst);

		/* Copy the overlays to the screen. */
		dst = topleft;
		SDL_RenderCopy(renderer, texture_overlay_rgb, NULL, &dst);
		dst = topright;
		SDL_RenderCopy(renderer, texture_overlay_rgba, NULL, &dst);
		dst = bottomleft;
		SDL_RenderCopy(renderer, texture_surface_rgb, NULL, &dst);
		dst = bottomright;
		SDL_RenderCopy(renderer, texture_surface_rgba, NULL, &dst);


		SDL_RenderPresent(renderer);

		/* Read the next frame into memory while the current one is being displayed. */
		if (player.bufferframe != (player.currentframe + 1) % videoinfo.numframes) {
			player.bufferframe = (player.currentframe + 1) % videoinfo.numframes;
			SDL_RWseek(videofile, videoinfo.frameoffsets[player.bufferframe], RW_SEEK_SET);
			SDL_RWread(videofile, player.buffer, 1, videoinfo.framesize);
		} else {
			SDL_Delay(1);
		}
	}

	return 0;
}

int main(int argc, char* argv[])
{
	int result;
	window = NULL;
	renderer = NULL;
	texture_yuv = NULL;
	texture_overlay_rgb = NULL;
	texture_overlay_rgba = NULL;
	texture_surface_rgb = NULL;
	texture_surface_rgba = NULL;
	texture_for_triangle = NULL;
	surface_rgb = NULL;
	surface_rgba = NULL;
	videofile = NULL;

	SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "best");

	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTS | SDL_INIT_TIMER) < 0) {
		SDL_Log("Could not initialize SDL");
		return 1;
	}

	result = LoadAndPlay(argc, argv);

	if (videofile != NULL)
		SDL_RWclose(videofile);

	if (surface_rgb != NULL)
		SDL_FreeSurface(surface_rgb);
	if (surface_rgba != NULL)
		SDL_FreeSurface(surface_rgba);

	if (texture_yuv != NULL)
		SDL_DestroyTexture(texture_yuv);
	if (texture_overlay_rgb != NULL)
		SDL_DestroyTexture(texture_overlay_rgb);
	if (texture_overlay_rgba != NULL)
		SDL_DestroyTexture(texture_overlay_rgba);
	if (texture_for_triangle != NULL)
		SDL_DestroyTexture(texture_for_triangle);
	if (texture_surface_rgb != NULL)
		SDL_DestroyTexture(texture_surface_rgb);
	if (texture_surface_rgba != NULL)
		SDL_DestroyTexture(texture_surface_rgba);

	if (renderer != NULL)
		SDL_DestroyRenderer(renderer);
	if (window != NULL)
		SDL_DestroyWindow(window);
	SDL_free(player.buffer);
	SDL_Quit();

	return result;
}

/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
